@@ -1,8 +1,5 @@
 # -*- coding: utf-8 -*-
 
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
 import os
 import unittest
 
@@ -10,12 +7,11 @@ from junitparser.cli import verify
 
 
 class Test_Cli(unittest.TestCase):
-
     def test_verify(self):
         files_expectedexitcodes = {
-            'data/jenkins.xml': 1,
-            'data/no_fails.xml': 0,
-            'data/normal.xml': 1,
+            "data/jenkins.xml": 1,
+            "data/no_fails.xml": 0,
+            "data/normal.xml": 1,
         }
         for file, expected_exitcode in files_expectedexitcodes.items():
             path = os.path.join(os.path.dirname(__file__), file)
