@@ -1,9 +1,5 @@
 # -*- coding: utf-8 -*-
 
-from __future__ import absolute_import
-from __future__ import unicode_literals
-from __future__ import with_statement
-
 import locale
 import unittest
 from copy import deepcopy
@@ -25,10 +21,29 @@ from junitparser import (
     Element,
 )
 
+
 try:
-    import itertools.izip as zip
+    from lxml import etree as expected_lxml_etree
+
+    has_lxml = True
 except ImportError:
-    pass
+    from xml.etree import ElementTree as expected_xml_etree
+
+    has_lxml = False
+
+
+class Test_XmlPackage(unittest.TestCase):
+    @unittest.skipIf(has_lxml, "xml package is used unless lxml is installed")
+    def test_xml_etree(self):
+        from junitparser.junitparser import etree as actual_etree
+
+        self.assertEqual(actual_etree, expected_xml_etree)
+
+    @unittest.skipUnless(has_lxml, "lxml package has to be installed")
+    def test_lxml_etree(self):
+        from junitparser.junitparser import etree as actual_etree
+
+        self.assertEqual(actual_etree, expected_lxml_etree)
 
 
 class Test_MergeSuiteCounts(unittest.TestCase):
@@ -483,6 +498,7 @@ class Test_TestSuite(unittest.TestCase):
         self.assertEqual(suite.errors, 1)
         self.assertEqual(suite.skipped, 1)
 
+
 class Test_TestCase(unittest.TestCase):
     def test_case_fromstring(self):
         text = """<testcase name="testname">
@@ -496,14 +512,34 @@ class Test_TestCase(unittest.TestCase):
         self.assertEqual(case.system_out, "System out")
         self.assertEqual(case.system_err, "System err")
 
-    def test_illegal_xml_multi_results(self):
+    def test_xml_multi_results(self):
         text = """<testcase name="testname">
         <failure message="failure message" type="FailureType"/>
         <skipped message="skipped message" type="FailureType"/>
         </testcase>
         """
         case = TestCase.fromstring(text)
-        self.assertRaises(JUnitXmlError)
+        # no assertion raised
+        self.assertEqual(case.name, "testname")
+        self.assertEqual(len(case.result), 2)
+
+    def test_multi_results(self):
+        case = TestCase("testname", "testclassname")
+        err = Error("err msg", "err_type")
+        fail1 = Failure("fail msg 1", "fail_type")
+        fail2 = Failure("fail msg 2", "fail_type")
+        fail3 = Failure("fail msg 3", "fail_type")
+        fail4 = Failure("fail msg 4", "fail_type")
+        case.result += [err]
+        self.assertEqual(len(case.result), 1)
+        case.result += [fail1]
+        self.assertEqual(len(case.result), 2)
+        case.result += [fail2]
+        self.assertEqual(len(case.result), 3)
+        case.result += [fail3]
+        self.assertEqual(len(case.result), 4)
+        case.result += [fail4]
+        self.assertEqual(len(case.result), 5)
 
     def test_case_attributes(self):
         case = TestCase()
@@ -731,7 +767,3 @@ class Test_Attrs(unittest.TestCase):
         self.assertEqual(element.text, "foo")
         self.assertEqual(element.int, 10)
         self.assertEqual(element.float, 8.5)
-
-
-if __name__ == "__main__":
-    unittest.main()
